#pragma once

#pragma comment(lib, "SUB2r-lib")

#ifndef _UNICODE
#error can only be compiled in UNICODE mode
#endif

#if !defined __STDCPP_THREADS__ || !defined _MT
#error the library is multi-threaded
#endif

#if _M_X64 != 100 || defined _M_IX86
#error must target x64-bit platform
#endif

#include <cstdint>
#include <chrono>
#include <memory>
#include <cstdint>
#include <string>
#include <vector>
#pragma warning(push)
#pragma warning(disable: 4996 26429 26440 26472 26496)
#include <gsl.h>
#pragma warning(pop)

#ifdef _WIN64
#define _WIN32_IE 0x0700

#include <SDKDDKVer.h>

#define WIN32_LEAN_AND_MEAN                 // Exclude rarely-used stuff from Windows headers
#define _ATL_CSTRING_EXPLICIT_CONSTRUCTORS  // some CString constructors will be explicit
#define _AFX_NO_MFC_CONTROLS_IN_DIALOGS     // remove support for MFC controls in dialogs

#ifndef VC_EXTRALEAN
#define VC_EXTRALEAN  // Exclude rarely-used stuff from Windows headers
#endif

#include <afx.h>

#include <mfidl.h>

#endif

using ULONG = unsigned long;

namespace S2R {

using namespace std;

class IImgSensor;
struct SUB2rCyFx3Device;

// create a span that represents a byte stream for a given object _x
template<class T> gsl::span<std::byte> byte_span(T& _x) noexcept { return { reinterpret_cast<std::byte*>(&_x), sizeof(T) }; }

#pragma region DebugLog

struct ApiCmdLog{
    using timestamp_t = decltype(chrono::high_resolution_clock::now());
    struct Entry{
        Entry(bool _dir2Dev, uint8_t _cmd, uint16_t _val, uint16_t _idx) noexcept
            : type{DebugLogType::fx3}
            , fx3{_dir2Dev, _cmd, _val, _idx}
        {}
        Entry(uint16_t _addr = 0, uint16_t _val = 0, uint16_t _mask = 0) noexcept
            : type{DebugLogType::i2c}
            , i2c{_addr, _val, _mask, 0}
        {}

        timestamp_t started{ chrono::high_resolution_clock::now() };
        chrono::nanoseconds duration{0ns};
        const enum class DebugLogType : uint8_t{fx3, i2c} type;
        bool success{};
        union{
            struct{
                const bool dir2Dev;
                const uint8_t cmd;
                const uint16_t val;
                const uint16_t idx;
                array<uint8_t, 4> ret{};   // first 4 bytes of the return buffer
            } fx3;
            struct{
                const uint16_t addr;
                const uint16_t val;
                const uint16_t mask;
                uint16_t valRead;
                uint16_t valWrite;
            } i2c;
        };
        [[nodiscard]] string toString() const;
        [[nodiscard]] static string legend();
    private:
        [[nodiscard]] string _toStrTime() const;
        [[nodiscard]] string _toStrData() const;
        [[nodiscard]] string _toStrDataFX3() const;
        [[nodiscard]] string _toStrDataI2C() const;
    };

    // add a new entry on construction and start the clock, update status by caller, log time in destructor
    struct Helper{
        Helper(bool _dir2Dev, uint8_t _cmd, uint16_t _val, uint16_t _idx): Helper{ ApiCmdLog::_add(_dir2Dev, _cmd, _val, _idx) }{}
        Helper(uint16_t _addr, uint16_t _val, uint16_t _mask)            : Helper{ ApiCmdLog::_add(_addr, _val, _mask) }        {}
        Helper() = delete;
        Helper(const Helper&) = default;
        Helper(Helper&&) = default;
        Helper& operator=(const Helper&) = delete;
        Helper& operator=(Helper&&) = delete;
        ~Helper(){ close(); }
        [[nodiscard]] auto & curr() noexcept { return ApiCmdLog::_isIdxValid(idx) ? ApiCmdLog::m_log[idx] : dummy; }
                      void close() noexcept;      // log the time and close the entry
    private:
        Helper(size_t _idx): active{ ApiCmdLog::_isIdxValid(_idx) }, idx{ _idx } {}
        const size_t idx;           // this entry's index in the log
        bool active{};              // should only change status or log time if `active == true`
		static Entry dummy;         // if the log is not enabled provide this to clients to avoid crashes
    };

    static void enable(bool _b)  noexcept { m_enabled = _b; }
    static void clear()          noexcept { m_log.clear(); }
    static const auto & getLog() noexcept { return m_log; }

protected:
    // add a new entry and return its index (or `(size_t)0 - 1` if log is not enabled)
    [[nodiscard]] static size_t _add(bool _dir2Dev, uint8_t _cmd, uint16_t _val, uint16_t _idx);
    [[nodiscard]] static size_t _add(uint16_t _addr, uint16_t _val, uint16_t _mask);
    [[nodiscard]] static bool _isIdxValid(size_t _idx) noexcept { return m_enabled && _idx < m_log.size(); }

private:
    inline static vector<Entry> m_log;
    inline static bool m_enabled{false};
    inline static const timestamp_t m_started{chrono::high_resolution_clock::now()};
};


#pragma endregion

#pragma region VersionInfo

// abstract interface for FX3 and FPGA version information, actual implementations are in the .cpp
// this class is also responsible for obtaining info on the latest available firmware for download
class VersionInfo
{
public:
    enum class ReleaseType : uint8_t { dev, alpha, beta, eval, rc, release, backport, ebf };
    //VersionInfo() = default;
    //VersionInfo(const VersionInfo&) = default;
    //VersionInfo(VersionInfo&&) = default;
    //VersionInfo & operator=(const VersionInfo&) = default;
    //VersionInfo & operator=(VersionInfo&&) = default;
    virtual ~VersionInfo() = default;

    // version info is returned from API as 4 bytes
	virtual void fromBuffer(gsl::span<std::byte> _buf = {}) = 0;

    // fields' descriptions for the one-liner
    [[nodiscard]] static wstring legend();

    // textual format for the release cycle type
    [[nodiscard]] static wstring relType(ReleaseType _rel);

    // total number of supported release types
    [[nodiscard]] static int countRelTypes() noexcept;

    // detailed line-by-line break-down
    [[nodiscard]] virtual wstring format(const wstring & _i) const = 0;

    // one-liner version info
    [[nodiscard]] virtual operator string() const = 0;

    // firmware upgrade info based on the specified earliest release cycle type
    struct FwInfo final{
        // parse data string into version information, _cgiVarName is used for the sanity check
        FwInfo(const string& _cgiVarName = {}, const string & _data = {});
        [[nodiscard]] bool operator<(const FwInfo& _o) const noexcept {return bld < _o.bld;}

        ReleaseType  rel{ReleaseType::rc};  // release type
        unsigned int bld{};                 // build #
        string       url;                   // where to get the latest code
        int          fsz{-1};               // file size
    };

    // Get the full list of matching (including downgrade) firmware files from the server.
    // Product ID, hardware config ID, and vendor ID from the version information must strictly match.
    // Throws an exception of type std::string on error.
    [[nodiscard]] virtual vector<FwInfo> listFw(ReleaseType _rel = ReleaseType::rc) const = 0;

    [[nodiscard]] virtual FwInfo latestFw(ReleaseType _rel = ReleaseType::rc) const = 0;

    // check if the given build# is an upgrade to the current code
    [[nodiscard]] virtual bool isUpgrade(unsigned int _bldNo) const noexcept = 0;

    // check if the build version is at least the given number or higher
    [[nodiscard]] virtual bool isAtLeast(unsigned int _bldNo) const noexcept = 0;
};

#pragma endregion

#pragma region UFix_8_8

// helper class for reading/writing unsigned fixed numbers in 2 8-bit bytes, where the MSB is the
// integer part, and the LSB is the fractional
class UFix_8_8 final{
public:
    constexpr UFix_8_8(uint8_t _i, uint8_t _f = 0) noexcept : m_valI(_i), m_valF(_f) {}
    constexpr UFix_8_8(double _d) noexcept;
    UFix_8_8(gsl::span<const std::byte> _buf) noexcept;   // from read buffer of 2 bytes

    [[nodiscard]] constexpr operator double() const noexcept {return m_valI + m_valF / 256.;}
    [[nodiscard]] constexpr operator uint16_t() const noexcept { return m_valI << 8 | m_valF; }   // make 2 bytes for writing to FX3

                  constexpr void operator+=(double _inc) noexcept;  // increment by _inc (positive or negative), abiding the bounds
    [[nodiscard]] constexpr uint8_t i() const noexcept {return m_valI;}
    [[nodiscard]] constexpr uint8_t f() const noexcept {return m_valF;}

private:
    uint8_t m_valI{}, m_valF{};
};

#pragma region UFix_8_8 inlines

inline UFix_8_8::UFix_8_8(gsl::span<const std::byte> _buf) noexcept
    : m_valI{static_cast<uint8_t>(_buf[0])}
    , m_valF{static_cast<uint8_t>(_buf[1])}
{
    assert(_buf.size_bytes() >= 2);
}

#pragma endregion

#pragma endregion

#pragma region FPGA configuration status

struct SPIConfigStatus final{
    SPIConfigStatus(const gsl::span<std::byte> _buf) noexcept{
        if(!_buf.empty()){
            m_ui8[0] = static_cast<uint8_t>(_buf[0]);
            m_ui8[1] = static_cast<uint8_t>(_buf[1]);
        }
    }

    constexpr SPIConfigStatus(uint16_t _val = 0) noexcept : m_ui16{_val} {}

    [[nodiscard]] constexpr operator uint16_t() const noexcept { return m_ui16; }

    [[nodiscard]] constexpr bool hasErrors() const noexcept; // true if any of the error bits is set

    union{
        uint16_t m_ui16{};
        uint8_t  m_ui8[2];
        struct{
            bool
                 configStarted      : 1
               , initializeOK       : 1
               , checkIdOK          : 1
               , eraseSwitchWordOK  : 1
               , eraseOK            : 1
               , programOK          : 1
               , verifyOK           : 1
               , programSwitchWordOK: 1
               , configNotBusy      : 1
               , configDone         : 1
               , configError        : 1
               , idCodeError        : 1
               , eraseError         : 1
               , programError       : 1
               , timeoutError       : 1
               , crcError           : 1
               ;
        } bit;
    };
};

#pragma region SPIConfigStatus inlines

constexpr inline bool S2R::SPIConfigStatus::hasErrors() const noexcept
{
    return bit.configError || bit.idCodeError || bit.eraseError || bit.programError || bit.timeoutError || bit.crcError;
}

#pragma endregion

#pragma endregion

#pragma region FX3

// FPGA access over Cypress USB (FX3) device
class FX3
{
public:
    // this is NOT a comprehensive list of supported commands, check the wiki for more
    enum class Fx3Cmd : uint8_t{
        check_bootloader  = 0xA0
      , calibrate_dpc     = 0xA2  // force the run of a DPC calibration
      , reconfig_fpga     = 0xA3  // causes FPGA to reconfigure itself from SPI Flash
      , i2c_bridge        = 0xA4  // access I2C (mainly the image chip)
      , raw_video         = 0xA5  // enable/disable "raw video mode", AKA "rev. 26", AKA "Cinema DNG alike"
      , erase_spi_flash   = 0xAA  // Any write to this location invalidates the FX3 SPI Flash and causes the FX3 to reset itself to a bootloader mode for reprogramming
      , fx3_version       = 0xAB
      , fpga_version      = 0xAC
      , fpga_config_ctrl  = 0xAD  // Any write to this location will put the FPGA into configuration mode. Only a reset of the FX3 and FPGA will restore normal operation
      , fpga_config_status= 0xAE
      , spi_write_enable  = 0xAF  // Reconfigure the FX3 IOMatrix to disable GPIF and enable SPI
      , fx3_reset         = 0xE0
    };

public:
    FX3(chrono::milliseconds _defaultTimeout = 500ms) noexcept;
    FX3(const FX3&) = delete;
    FX3(FX3&&) = default;
    FX3 & operator=(const FX3&) = delete;
    FX3 & operator=(FX3&&) = default;
    virtual ~FX3();

    [[nodiscard]] bool isValid() const;     // if true - can send commands to this device
    [[nodiscard]] uint8_t deviceCount() const noexcept; // how many Cypress devices are found attached to the system
            bool open(uint8_t _dev);        // open device # _dev, which is a 0-based index
    virtual void close() noexcept;          // close currently opened device, invalidating its handler
    // device path is unique per device and is valid across reboots
    // returns empty string if device not found
    [[nodiscard]] static wstring getDevPath(uint8_t _devNo = 0);
#ifdef _WIN32
    bool open(IMFActivate * _src);          // open FX3 device that corresponds to the given IUnknown (IMFMediaSource or IMFActivate)
    [[nodiscard]] CComPtr<IUnknown> toInterface() const;  // find the media interface for this FX3 device
#endif

    // this family of operator[] functions provides read-only access to "properties" of the device
    enum class PropBool   {h_speed, s_speed, bus_powered};
    enum class PropString {protocol};
    enum class PropWString{dev_name, friendly_name, dev_path, manufacturer, product_name};
    enum class PropUint16 {vendor_id, product_id};
    enum class PropInt    {usb_address, serial_no, max_bus_power};
    enum class PropPair   {host_version, protocol_version, usb_class};
    [[nodiscard]] bool           operator[](PropBool    _prop) const;
    [[nodiscard]] string         operator[](PropString  _prop) const;
    [[nodiscard]] wstring        operator[](PropWString _prop) const;
    [[nodiscard]] uint16_t       operator[](PropUint16  _prop) const;
    [[nodiscard]] int            operator[](PropInt     _prop) const;
    [[nodiscard]] pair<int, int> operator[](PropPair    _prop) const;

    // H/W version info
    [[nodiscard]] const VersionInfo & fx3Version() const noexcept;
    [[nodiscard]] const VersionInfo & fpgaVersion() const noexcept;

    // run a vendor request command
    enum class VrCmdOpType {read, write};
	bool vrCmd(uint8_t _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, gsl::span<std::byte> _buf = {}) const;
    bool vrCmd(Fx3Cmd  _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, gsl::span<std::byte> _buf = {}) const;
    template<typename T = enable_if_t<is_standard_layout<T>::value> >
    bool vrCmd(uint8_t _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, T & _valBuf) const;
    template<typename T = enable_if_t<is_standard_layout<T>::value> >
    bool vrCmd(Fx3Cmd  _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, T & _valBuf) const;

    // program the FPGA with the provided buffer, fail if timeout is reached
    [[nodiscard]] bool programFwFPGA(const vector<std::byte> & _buf, chrono::milliseconds _timeout = 90s) const;

    // program the FX3 SPI flash with the provided buffer
    [[nodiscard]] bool programFwFX3(const vector<std::byte> & _buf) const;

    // utility functions (basically just shortcuts)
    bool stopStreaming() const;
    bool setLED(uint8_t _r, uint8_t _g, uint8_t _b) const;

protected:
    unique_ptr<SUB2rCyFx3Device> m_USBDevice;
    unique_ptr<VersionInfo>      m_fx3Ver;
    unique_ptr<VersionInfo>      m_fpgaVer;

protected:
    [[nodiscard]] virtual bool _onOpen();  // called from open() on success

private:
    void _readVersionInfo(Fx3Cmd _cmd, VersionInfo & _vi);
};

#pragma region FX3 inlines

inline const VersionInfo & FX3::fx3Version() const noexcept
{
    return *(m_fx3Ver.get());
}

inline const VersionInfo & FX3::fpgaVersion() const noexcept
{
    return *(m_fpgaVer.get());
}

inline bool FX3::vrCmd(Fx3Cmd _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, gsl::span<std::byte> _buf) const
{
    return vrCmd(static_cast<uint8_t>(_cmd), _opType, _val, _idx, _buf);
}

template<typename T>
inline bool S2R::FX3::vrCmd(uint8_t _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, T & _valBuf) const
{
    return vrCmd(static_cast<uint8_t>(_cmd), _opType, _val, _idx, byte_span(_valBuf));
}

template<typename T>
inline bool S2R::FX3::vrCmd(Fx3Cmd _cmd, VrCmdOpType _opType, uint16_t _val, uint16_t _idx, T & _valBuf) const
{
    return vrCmd(static_cast<uint8_t>(_cmd), _opType, _val, _idx, byte_span(_valBuf));
}
#pragma endregion

#pragma endregion end FX3

#pragma region Color Grading helper

// Color Grading helper to simplify access to related functionality
struct CG final{
    enum table : unsigned char{huevhue, huevsat, huevluma};
    struct ctrl{
        union{
            struct{
                bool bulk : 1;           // false for normal, true for bulk
                unsigned char idxMsb : 4;// 4 MSB of the "index"
                enum table tbl: 3;       // specific table
            };
            unsigned char all;
        };
        operator uint8_t() const { return all; }
    };
    static_assert(sizeof(ctrl) == 1);

    CG(FX3 & _fx3): m_fx3{_fx3}{}
    [[nodiscard]] operator bool() const { return !m_error; }         // check the state of the last operation
    [[nodiscard]] float read(table _tbl, uint16_t _idx);             // read a single value, setting error flag if failed
    [[nodiscard]] bool write(table _tbl, uint16_t _idx, float _val); // write a single value, returning false and setting error flag if failed

    // read/write values in a bulk burst, returning a vector of successfully read elements
    // if bulk is not supported - gracefully fallback on sequential access
    [[nodiscard]] vector<float> readBulk(table _tbl, uint16_t _idx0, int _n);
    // returns number of seccessully written values
    [[nodiscard]] uint16_t writeBulk(table _tbl, uint16_t _idx0, int _n, const vector<float> & _arr);

    [[nodiscard]] bool supported(table _tbl) const; // check if a given transformation is available

private:
    [[nodiscard]] bool _supported(bool _bulk, table _tbl) const; // check if a given transformation is available
    [[nodiscard]] float _f(table _tbl, int16_t _val);  // convert from device to user
    [[nodiscard]] uint16_t _i(table _tbl, float _val); // convert from user to device
    // setup the control block information for the subsequent read/write operation(s)
    [[nodiscard]] bool _setCtrl(bool _bulk, table _tbl, uint16_t _idx);
    // read a signed int value of _bits width from a pair of registers, set m_error on error
    [[nodiscard]] int16_t _read(uint8_t _bits);
    // write an unsigned int value into MSB/LSB registers
    [[nodiscard]] bool _write(uint16_t _val);

private:
    FX3 & m_fx3;
    bool m_error{false}; // result of the last operation
    static constexpr auto fx3CmdI2C{S2R::FX3::Fx3Cmd::i2c_bridge};
    static constexpr auto fx3CmdRead{S2R::FX3::VrCmdOpType::read};
    static constexpr auto fx3CmdWrite{S2R::FX3::VrCmdOpType::write};
};

#pragma endregion end Color Grading helper

#pragma region I2C

using CVecRegValPairs = vector<pair<uint16_t, uint16_t>>;
using CVecRegAddr = vector<uint16_t>;

// allows access to image sensor (currently OV) chipset via its I2C interface over Cypress USB (FX3) device
class I2C final : public FX3
{
public:
    I2C() noexcept;

    [[nodiscard]] operator bool() const;

                  void    setCtrlPort(uint8_t _port);    // set control port to a non-standard one
    [[nodiscard]] uint8_t getCtrlPort() const noexcept;  // a control port is usually set at open()

    // get appropriate interface class for the detected chipset, almost a factory method
    [[nodiscard]] const IImgSensor * operator->() const noexcept;
    [[nodiscard]] const IImgSensor * sensor()     const noexcept;
    [[nodiscard]]       IImgSensor * sensor()           noexcept;

    // Primary interface to send individual "raw" I2C commands.
    // Performs a read if _mask is 0x00, write otherwise.
    // When writing only affects the bits where mask is set to 1.
    [[nodiscard]] uint16_t operator()(uint16_t _addr, uint16_t _val = 0, uint16_t _mask = 0) const;

    // (blindly) write a group of registers, optionally at a VSYNC boundary.
    // If _bUseGroupWrite == true will use the group #0 (will overwrite it)
    void groupWrite(const CVecRegValPairs & _group, bool _bUseGroupWrite = true) const;

    // read/write N bits into a set of registers, first register gets the "tail" MSB bits with the last
    // register getting the LSB of the value into its bits, starting the fill with the MSB
    [[nodiscard]] int  readN(const CVecRegAddr & _r, int _n) const;
                  void writeN(const CVecRegAddr & _r, int _n, int _val, bool _bUseGroupWrite = true) const;

protected:
    [[nodiscard]] bool _onOpen() override;

    [[nodiscard]] uint16_t _cmd(uint16_t _addr, uint16_t _val, uint16_t _mask = 0) const;

    // split a _val into registers _r based on number of bits _n
    // this will read the partial byte and adjust the stored value to preserve the "extra" bits
    [[nodiscard]] CVecRegValPairs _splitN(const CVecRegAddr & _r, int _n, int _val) const;

    // set and clear bits of _val to _bits, but only affect those where the _mask is "1"
    [[nodiscard]] static constexpr uint16_t _bitsSetAndClear(uint16_t _val, uint16_t _bits, uint16_t _mask) noexcept;

private:
    uint8_t m_ctrlPort{0x20};
    mutable unique_ptr<IImgSensor> m_sensor;
};

#pragma region I2C inlines

inline uint8_t I2C::getCtrlPort() const noexcept
{
    return m_ctrlPort;
}

inline const IImgSensor * I2C::operator->() const noexcept
{
    return m_sensor.get();
}

inline const IImgSensor * I2C::sensor() const noexcept
{
    return m_sensor.get();
}

inline IImgSensor * I2C::sensor() noexcept
{
    return m_sensor.get();
}

inline uint16_t I2C::operator()(uint16_t _addr, uint16_t _val, uint16_t _mask) const
{
    return _cmd(_addr, _val, _mask);
}

inline constexpr uint16_t I2C::_bitsSetAndClear(const uint16_t _val, const uint16_t _bits, const uint16_t _mask) noexcept
{
    // bit + mask + value = result
    // 0 + 0 + 0 = 0
    // 0 + 0 + 1 = 1
    // 0 + 1 + 0 = 0
    // 0 + 1 + 1 = 0	<-- update
    // 1 + 0 + 0 = 0
    // 1 + 0 + 1 = 1
    // 1 + 1 + 0 = 1	<-- update
    // 1 + 1 + 1 = 1
    return _bits & _mask | _val & ~_mask;
}

#pragma endregion inlines

#pragma endregion I2C

#pragma region IImgSensor

struct TestPattern final{
    TestPattern() = default;
    TestPattern(int _bits) noexcept;
    operator uint16_t() const noexcept;

    // READ mode is used to read the current value from the chip
    enum {read = -1, color_bar = 0, random = 1, square = 2, black = 3}
        m_mode{read};

    enum class ColorBar {std = 0, tb_dark = 1, rl_dark = 2, bt_dark = 3}
        m_colorBarType{ColorBar::std};

    enum class SqColor {clr = 0, bw = 1}
        m_squareType{SqColor::clr};

    bool m_transparent{};
    bool m_rollingbar {};
    bool m_enabled    {};
};

// interface provider, knows which commands do what for a specific chipset
class IImgSensor {
protected:
    // internal helper class to allow code like the following:
    // m_sensor[reg_val_name] = 123;
    // and even
    // m_sensor[rvn1] = m_sensor[rvn2] = 123;
    struct CDevProp final{
        CDevProp()                              = delete;   // force the use of initializer list
        CDevProp& operator=(const CDevProp& _o) = delete;   // force the operator int() for assignments
        CDevProp& operator=(int _val);
        CDevProp& operator+=(int _inc);
        CDevProp& operator-=(int _dec);
        CDevProp& operator*=(double _mul);
        CDevProp& operator/=(double _div);
        [[nodiscard]] operator int() const;

        const I2C&        m_fx3;
        const uint8_t     m_nBits;
        const CVecRegAddr m_addr;
        const bool        m_atomic;
    };

public:
    [[nodiscard]] static unique_ptr<IImgSensor> create(const I2C & _fx3);   // factory
    IImgSensor() = delete;
    IImgSensor(const IImgSensor&) = default;
    IImgSensor(IImgSensor&&) = default;
    IImgSensor & operator=(const IImgSensor&) = delete;
    IImgSensor & operator=(IImgSensor&&) = delete;
    virtual ~IImgSensor() = default;

    enum class Chipset{unknown, OV9714, OV9750, OV10823, OV13850, OV2770};
    enum class Value{gain_r, gain_g, gain_b, gain_global, exposure, black_level, tp_seed};
    enum class Mode{standby, flip_h, flip_v, manual_wb, manual_gain, manual_exposure, manual_tp_seed};
    enum class Command{reset, color_correct_for_flips, fsin};
    enum class GroupWriteMode{immediate, at_line_sync, at_vert_sync};

    [[nodiscard]] Chipset chipsetType()  const noexcept;
    [[nodiscard]] string  chipsetName()  const;
    [[nodiscard]] string  chipsetIdStr() const;
    // empty chipset is "compatible" with any other chipset
    [[nodiscard]] virtual bool compatible(string_view _cs) const;

    // chipset-specific limits for various values (gain, exposure, etc)
    [[nodiscard]] virtual int getLimit(Value _lt) const noexcept;
    virtual void cmd(Command _c)     const;
    // mode() and val() return the valid read value when reading, 0 or unspecified when writing
    virtual bool        mode(Mode _mt, int = -1) const;  // set the new mode if it is >= 0
    virtual int         val(Value _vt, int = -1) const;  // set the new value if it is >= 0
            int         operator[](Value _vt)    const;
    [[nodiscard]] virtual CDevProp    operator[](Value _vt) = 0;
    virtual TestPattern testPattern(TestPattern _tp = {}) const;

    // support for OmniVision's group register writes
    virtual void groupStart(uint8_t _grpNo) const = 0;
    virtual void groupEnd(uint8_t _grpNo, GroupWriteMode _mode = GroupWriteMode::at_vert_sync) const = 0;

protected:
    IImgSensor(const I2C & _fx3, Chipset _cs = Chipset::unknown) noexcept;
    void _notImplementedMsg() const;

protected:
    const I2C &   m_fx3;
    const Chipset m_chipset;
};

#pragma region IImgSensor inlines

inline IImgSensor::CDevProp & IImgSensor::CDevProp::operator=(int _val)
{
    if(m_nBits > 0){
        m_fx3.writeN(m_addr, m_nBits, _val, m_atomic);
    }
    return *this;
}

inline IImgSensor::CDevProp & IImgSensor::CDevProp::operator+=(int _inc)
{
    const auto value = m_fx3.readN(m_addr, m_nBits);
    if(value >= 0){
        const auto newVal = value + _inc;
        if(newVal >= 0){
            m_fx3.writeN(m_addr, m_nBits, newVal, m_atomic);
        }
    }
    return *this;
}

inline IImgSensor::CDevProp & IImgSensor::CDevProp::operator-=(int _dec)
{
    const auto value = m_fx3.readN(m_addr, m_nBits);
    if(value >= 0){
        const auto newVal = value - _dec;
        if(newVal >= 0){
            m_fx3.writeN(m_addr, m_nBits, newVal, m_atomic);
        }
    }
    return *this;
}

inline IImgSensor::CDevProp & IImgSensor::CDevProp::operator*=(double _mul)
{
    const auto value = m_fx3.readN(m_addr, m_nBits);
    if(value >= 0){
        m_fx3.writeN(m_addr, m_nBits, static_cast<int>(value * _mul), m_atomic);
    }
    return *this;
}

inline IImgSensor::CDevProp & IImgSensor::CDevProp::operator/=(double _div)
{
    const auto value = m_fx3.readN(m_addr, m_nBits);
    if(value >= 0){
        m_fx3.writeN(m_addr, m_nBits, static_cast<int>(value / _div), m_atomic);
    }
    return *this;
}

inline IImgSensor::CDevProp::operator int() const
{
    return m_fx3.readN(m_addr, m_nBits);
}

inline IImgSensor::Chipset IImgSensor::chipsetType() const noexcept
{
    return m_chipset;
}

inline int IImgSensor::operator[](Value _vt) const
{
    return val(_vt, -1);
}

inline IImgSensor::IImgSensor(const I2C & _fx3, Chipset _cs) noexcept
    : m_fx3(_fx3)
    , m_chipset(_cs)
{
}

#pragma endregion inlines

#pragma endregion IImgSensor

} // namespace S2R
